/*
 * =====================================================================================
 * 
 *        Filename:  rsh-fifo.c
 * 
 *     Description:  The 'rsh' client fifo.
 * 
 *         Version:  $Id$
 *         Created:  28.12.2004 09:03:19 CET
 * 
 *          Author:  Karel Zak <kzak@redhat.com>
 * 
 * =====================================================================================
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <fcntl.h>
#include <pwd.h>
#include <string.h>
#include <signal.h>
#include <errno.h>

#define BUFSIZE		PIPE_BUF
#define MAXINTSTR	16
#define MAXCMDSZ	BUFSIZE
#define MAXDATASZ	BUFSIZE

#define PATH_RSH "/usr/bin/rsh"
#define PATH_FIFO "/tmp/rsh-%s-%s-%s"	/* s1=host, s2=username, s3=postfix */

#define MASTER_NAME	"master"

static char *host, *user, *prog;
static pid_t chpid;
static int chend, pipend;


#define CMD_HEADERSZ	(2*sizeof(int))
#define DATA_HEADERSZ	(1+sizeof(int))

#define CMD_MARK_END	"rsh-fifo-end"
#define CMD_MARK_MAXLEN	(sizeof(CMD_MARK_END))

#define CMD_ECHO_END	"; echo \"" CMD_MARK_END "\"; echo \"" CMD_MARK_END "\" 1>&2\n"
#define CMD_ECHO_MAXLEN	(sizeof(CMD_ECHO_END))

static void
usage()
{
	printf(	"\n%s host [-l login] [-C | -D | -R command]\n\n"
		"    -h             - this help\n"
		"    -C             - creates persistent connection to rsh server\n"
		"    -D             - destroys connection to rsh server\n"
		"    -R command     - sends command to remote shell\n\n", prog);
	exit(EXIT_FAILURE);
}

static void
failure(char *msg)
{
	if (errno)
		fprintf(stderr, "%s: %s: %s\n", prog, msg, strerror(errno));
	else
		fprintf(stderr, "%s: %s\n", prog, msg);
	exit(EXIT_FAILURE);
}


/* child process is very simple -- it starts 'rsh' client only */
static int
rsh()
{
	char *argv[7];
	int i=0;

	argv[i++] = PATH_RSH;
	argv[i++] = "-l";
	argv[i++] = user;
	argv[i++] = host;
	argv[i++] = "$SHELL";		/* TODO: use option for this */
	argv[i++] = NULL;
	
	execv(PATH_RSH, argv);
	failure("exec rsh client");
}

static char *
fifoname(char *buf, int bufsiz, char *postfix)
{
	snprintf(buf, bufsiz, PATH_FIFO, host, user, postfix);
	return buf;
}

static void
fifo_broker_cleanup(void)
{
	char fname[MAXPATHLEN];

	if (chpid)
		kill(chpid, SIGKILL);     /* my sweet child, you have to die ... */
	
	fifoname(fname, MAXPATHLEN, MASTER_NAME);
	unlink(fname);
}

static void
pipe_handler(int x)
{
//	fprintf(stderr, "signal %d\n", x);
	pipend = 1;
	
}

static void
command_write(int fd, int ident, char *cmd)
{
	char hdr[CMD_HEADERSZ], *p=hdr;
	int sz=strlen(cmd);
	
	if (sz > MAXCMDSZ)
		failure("unexpected command size");
		
	memcpy(p, &ident, sizeof(int));		/* 1.1) ident */
	p+=sizeof(int);
	memcpy(p, &sz, sizeof(int));		/* 1.2) size of command */
	p+=sizeof(int);

	write(fd, hdr, CMD_HEADERSZ);		/* 1) header */
	write(fd, cmd, sz);			/* 2) command */
}	

static int
command_read(int fd, int *ident, char *cmd, int *sz)
{
	char hdr[CMD_HEADERSZ];
	
	*sz=0;
	
	if (read(fd, hdr, CMD_HEADERSZ) < CMD_HEADERSZ)
		/* this happen if we wait for fifo client end */
		return 0;
	
	memcpy(ident, hdr, sizeof(int));		/* 1.1) ident */
	memcpy(sz, hdr+sizeof(int), sizeof(int));	/* 1.2) size of command */

	if (!*sz)
		return 0;
	if (*sz > MAXCMDSZ)
		failure("unexpected command size");

	if (read(fd, cmd, *sz) < *sz)
		failure("read command");

	return 1;
}	

static void
data_write(int fd, char type, char *buf, int sz)
{
	char hdr[DATA_HEADERSZ];
	
	if (sz > MAXDATASZ)
		failure("unexpected data size");

	*hdr=type;					/* 1.1) type */
	memcpy(hdr+1, &sz, sizeof(int)); 		/* 1.2) size of data */
	
	if (write(fd, hdr, DATA_HEADERSZ) < DATA_HEADERSZ)	/* 1) header */
		return;
	if (!sz)
		return;
	
	write(fd, buf, sz);				/* 2) data */
}

static int
data_read(int fd, char *type, char *buf, int *sz)
{
	char hdr[DATA_HEADERSZ];
	
	if (read(fd, hdr, DATA_HEADERSZ) < DATA_HEADERSZ)	/* 1) header */
		return 0;
	
	*type = *hdr;				/* 1.1) type */
	memcpy(sz, hdr+1, sizeof(int));		/* 1.2) size of data */
	
	if (!*sz)
		return 1;

	if (*sz > MAXDATASZ)
		failure("unexpected data size");

	if (read(fd, buf, *sz) < *sz)
		failure("read data");

	return 1;
}

static int
rsh_select(fd_set *set, int out, int err)
{
	int n=-1, sel;
	
	FD_ZERO(set);

	FD_SET(out, set);	/* rsh 'stdout' */
	n = n > out ? n : out;
	
	FD_SET(err, set);	/* rsh 'stderr' */
	n = n > err ? n : err;

//	fprintf(stderr, "SELECT RSH\n");
	sel=select(n+1, set, NULL, NULL, NULL);
	
	if (chend)
		/* rsh client ends */
		exit(EXIT_SUCCESS);
	
	if (sel==-1 && errno!=EINTR)
		failure("select");
	return sel;
}

static int
open_slave_fifo(int ident, int flag)
{
	int fd;
	char x[ MAXINTSTR ];
	char fname[MAXPATHLEN];

	snprintf(x, MAXINTSTR, "%d", ident);
	fifoname(fname, MAXPATHLEN, x);

//	fprintf(stderr, "OPEN %s\n", fname);
	while ((fd = open(fname, flag))==-1);	/* wait until it's possible open */

	return fd;
}

/* this is client of fifo_broker() */
static int
sendcmd(char *cmd)
{
	int master, slave, sz, ident = (int) getpid();
	char buf[MAXDATASZ], type;
	char fname[MAXPATHLEN];

	signal(SIGPIPE, pipe_handler);

	/* open master -- this unblock backend process */
	fifoname(fname, sizeof(fname), MASTER_NAME);
	if ((master=open(fname, O_WRONLY))==-1)
		failure("open master fifo");

	/* send command to backend */
	command_write(master, ident, cmd);

	/* open slave */
	snprintf(buf, MAXINTSTR, "%d", ident);
	fifoname(fname, sizeof(fname), buf);
	unlink(fname);

//	fprintf(stderr, "CREATE-SLAVE: %s\n", fname);
	if (mkfifo(fname, S_IRUSR | S_IWUSR))
		failure(fname);
	
	/* this call wait until backend process open fifo for read */
	if ((slave = open(fname, O_RDONLY))==-1)
		failure("open slave fifo");

	for(;;) 
	{
		if (!data_read(slave, &type, buf, &sz))
			break;
		if (type=='X')
			break;
		if (sz)
			write(type=='E' ? 2 : 1, buf, sz);
	}
	close(slave);
	close(master);

	snprintf(buf, MAXINTSTR, "%d", ident);
	fifoname(fname, MAXPATHLEN, buf);
//	fprintf(stderr, "DESTROY-SLAVE: %s\n", fname);
	unlink(fname);

	return EXIT_SUCCESS;
}

/* This backend for sendcmd() 
 */
static int
fifo_broker(int in, int out, int err)
{
	char fname[MAXPATHLEN], *eend=NULL, *oend=NULL;
	char buf[MAXDATASZ];
	int master=-1, slave=-1, ident=0;
	
	atexit(fifo_broker_cleanup);
	signal(SIGPIPE, pipe_handler);
	
	/* create backend fifo */
	fifoname(fname, MAXPATHLEN, MASTER_NAME);
	unlink(fname);

	if (mkfifo(fname, S_IRUSR | S_IWUSR))
		failure(fname);
	
	for(;;)
	{
		fd_set set;
		int cc;

		errno = 0;
//		fprintf(stderr, "master loop\n");
		
		if (!ident)
		{
			if (slave!=-1)
			{
//				fprintf(stderr, "CLOSE SLAVE\n");
				if (!pipend)
					data_write(slave, 'X', NULL, 0);
				close(slave);
				slave = -1;
			}
			/* wait for new client */
			if (master!=-1)
				close(master);

			if ((master = open(fname, O_RDONLY))==-1)
				failure("open master fifo");
			
//			fprintf(stderr, "NEW-CLIENT?\n");
			
			if (!command_read(master, &ident, buf, &cc))
				continue;
			
//			fprintf(stderr, "IDENT=%d COMMAND(%d)=>>%s<<\n",  ident, cc, buf);
			
			if (cc==0 || ident==0)
			{
				ident=0;
				continue;
			}
			/* TODO: add to failure() check -- if slave already open
			 * send 'X' and close it 
			 */
			if (!(slave = open_slave_fifo(ident, O_WRONLY)))
			{
				ident=0;
				continue;
			}
			
			/* send command to rsh */
			if (write(in, buf, cc) < cc)
				failure("write data to rsh stdin");
			
			/* mark end of command */
			if (write(in, CMD_ECHO_END, sizeof(CMD_ECHO_END)) < 0)
				failure("write mark to rsh stdin");
		}
		
		if ((eend && oend) || pipend)
		{
//			fprintf(stderr, "RESET\n");
			eend = oend = NULL;
			pipend = ident = 0;
			continue;
		}

		/* wait for 'rsh' result */
		if (!rsh_select(&set, out, err))
			break;
		
		memset(buf, 0, sizeof(buf));
		
		/* rsh stdout active */
		if (FD_ISSET(out, &set))
		{
//			fprintf(stderr, "RSH-OUT active\n");
			while(oend==NULL && (cc=read(out, buf, sizeof(buf))) > 0)
			{
//				fprintf(stderr, "RSH-OUT(%d): >>>%s<<\n", cc, buf);
				if (cc < sizeof(buf))
				{
					buf[cc+1] = '\0';
					if ((oend = strstr(buf, CMD_MARK_END)))
					{
						*oend = '\0';
						cc = oend - buf;
//						fprintf(stderr, "RSH-OUT: end\n");
					}
				}
//				fprintf(stderr, "RSH-OUT(%d): >>>%s<<\n", cc, buf);	
				if (pipend==0 && cc)
					data_write(slave, 'O', buf, cc);
			}
		}

		/* rsh stderr active */
		if (FD_ISSET(err, &set))
		{
//			fprintf(stderr, "RSH-ERR active\n");
			while(eend==NULL && (cc=read(err, buf, sizeof(buf))) > 0)
			{
//				fprintf(stderr, "RSH-ERR read\n");
				if (cc < sizeof(buf))
				{
					buf[cc+1] = '\0';
					if ((eend = strstr(buf, CMD_MARK_END)))
					{
						*eend = '\0';
						cc = eend - buf;
//						fprintf(stderr, "RSH-ERR: end\n");
					}
				}
				if (pipend==0 && cc)
					data_write(slave, 'E', buf, cc);
			}
		}
	}
//	fprintf(stderr, "standard end\n");

	if (slave!=-1)
	{
		if (!pipend)
			data_write(slave, 'X', NULL, 0);
	}
	exit(EXIT_SUCCESS);
}


#define set_std(p, wanted, dest) \
	do { \
		close(dest);               /* close original dest */                 \
		close(p[ ~wanted & 1 ]);   /* close non-wanted part of pipe */       \
		dup2(p[wanted], dest);     /* set 'dest' as wanted part of pipe */   \
		close(p[wanted]);                                                    \
	} while(0)	
				
#define P_READ	0		/* pipe[0] */
#define P_WRITE	1		/* pipe[1] */

static void
connect_sigchld_handler(int x)
{
	chend = 1;
	chpid = 0;
}

static int
connect()
{
	int perr[2], pout[2], pin[2];
	
	if (pipe(perr)!=0) 
		failure("pipe (err)");
	if (pipe(pin)!=0)
		failure("pipe (in)");
	if (pipe(pout)!=0) 
		failure("pipe (out)");
	
	signal(SIGCHLD, connect_sigchld_handler);
	
	switch((chpid = fork()))
	{
		case -1:
			failure("fork");
			break;
		case 0:
			signal(SIGCHLD, SIG_DFL);
			set_std(pin, P_READ, 0);	/* stdin */
			set_std(pout, P_WRITE, 1);	/* stdout */
			set_std(perr, P_WRITE, 2);	/* stderr */
			rsh();
			break;
		default:
			close(pin[P_READ]);
			close(pout[P_WRITE]);
			close(perr[P_WRITE]);
			fifo_broker(pin[P_WRITE], pout[P_READ], perr[P_READ]);
			break;
	}
	
	return EXIT_FAILURE;	/* never .. */
}



#define RSH_CONNECT 	1 << 1
#define RSH_DISCONNECT	1 << 2
#define RSH_RUN		1 << 3

int
main(int argc, char **argv)
{
	int ch, flag=0;
	char *command = NULL;
	
	prog = argv[0];

	/* fisrt arg is hostname */
	if (argc < 2 || *argv[1] == '-')
		usage();
	
	host = argv[1];
		
	while ((ch = getopt(argc-1, argv+1, "hl:CDR:")) != EOF)
	{
		switch(ch)
		{
			case 'l':
				user = optarg;
				break;
			case 'C':
				flag |= RSH_CONNECT;
			       	break;
			case 'D':
				flag |= RSH_DISCONNECT;
				break;
			case 'R':
				flag |= RSH_RUN;
				command = optarg;
				break;
			case 'h':
			default:
				usage();
		}
	}

	if (!user)
	{
		struct passwd *pw;
		errno = 0;
		
		if (!(pw = getpwuid(getuid())))
		{
			if (errno)
				perror(prog);
			else
				fprintf(stderr, "%s: failed load password fail entry about acutual user", prog);
			exit(EXIT_FAILURE);
		}
		user = pw->pw_name;
	}		
	
	if (flag & RSH_CONNECT)
	{
		if ((flag & RSH_DISCONNECT) || (flag & RSH_RUN))
			usage();
		exit( connect() );
	}
	if (flag & RSH_DISCONNECT)
	{
		if ((flag & RSH_CONNECT) || (flag & RSH_RUN))
			usage();
		exit( sendcmd("exit") );
	}
	if (flag & RSH_RUN)
	{
		if ((flag & RSH_CONNECT) || (flag & RSH_DISCONNECT) || command==NULL)
			usage();
		exit( sendcmd(command) );
	}
	usage();
	exit(EXIT_FAILURE);
}


